#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include "rpi.h"
#include "pilot.h"
#include "tick.h"
#include "input.h"

// ============================================================================
// Debounce time (20ms) in jiffie count
#define DEBOUNCE (20*HZ/1000)
// Long press if > 1s
#define LONG_PRESS_TIME (HZ)

// ============================================================================

struct GpioRegs {
  unsigned gppud;
  unsigned gppudclk[2];
};
typedef enum { EVT_INPUT_CHANGED, EVT_TICK } InputEvent;

typedef struct Input Input;

struct Input {
  const HwInput *hw;
  enum { IDLE, PRESSED, LONG_PRESSED } state;
  unsigned lastTime;
  // Next input for button list
  Input *next;
  struct {
    unsigned pressed, longpressed;
    unsigned jitterNb, jitterTime, jitterMax;
  } stats;
};

static Input *inList = NULL;

// ============================================================================

#define JITTER_ADD(t, i) {    \
  i->stats.jitterNb++;        \
  i->stats.jitterTime += t;   \
  if (t > i->stats.jitterMax) \
    i->stats.jitterMax = t;   \
}

#define PRESSED_INC(i)     (i->stats.pressed++)
#define LONGPRESSED_INC(i) { i->stats.longpressed++; i->stats.pressed++; }

static int inPrintStatsForInput(char *buf, size_t size, Input *inp) {
  int nb;
  unsigned val, jAvg = 0, jMax = 0;

  if (inp == NULL) return 0;

  if (inp->stats.jitterNb) {
    if (inp->stats.jitterTime) {
      jAvg = (inp->stats.jitterTime/inp->stats.jitterNb)*1000u/HZ;
      jMax = inp->stats.jitterMax*1000u/HZ;
    } else {
      jAvg = jMax = 1;
    }
  }

  val = gpio_get_value(inp->hw->gpio.pin);
  if (inp->hw->gpio.polarity == ACTIVE_LOW) val = (val ? 0 : 1);

  nb = inPrintStatsForInput(buf, size, inp->next);
  nb += scnprintf(buf+nb, size-nb, "  %-5s: %s activated %u (%u long), bounce:%u (avg:%ums, max:%ums)\n",
      inp->hw->name, val ? "active" : "idle  ", inp->stats.pressed, inp->stats.longpressed,
      inp->stats.jitterNb, jAvg, jMax);
  return nb;
}

int inPrintStats(char *buf, size_t size) {
  int nb = scnprintf(buf, size, "Inputs:\n");
  return nb+inPrintStatsForInput(buf+nb, size-nb, inList);
}

static void inResetStatsForInput(Input *inp) {
  inp->stats.pressed = inp->stats.longpressed = inp->stats.jitterNb = inp->stats.jitterTime = inp->stats.jitterMax = 0;
}

void inResetStats(void) {
  Input *ptr = inList;

  while (ptr) {
    inResetStatsForInput(ptr);
    ptr = ptr->next;
  }
}

// ============================================================================

static void inSendEvent(Input *input, InputEvent evt) {
  //printk(KERN_INFO "Tobe-Pilot: State Machine [%s:%d] <= %s (%lu).\n", input->hw->name, input->state, evt == EVT_TICK ? "Tick" : "Input_irq", jiffies);

  switch (input->state) {
    case IDLE:
      if (evt == EVT_INPUT_CHANGED) {
        if (jiffies - input->lastTime < DEBOUNCE) {
          JITTER_ADD(jiffies - input->lastTime, input);
          break; // ignoring glitch
        }
        input->lastTime = jiffies;
        input->state = PRESSED;
        if (input->hw->longPressed == NULL)
          inScriptRun(input->hw->pressed, 0); // do not wait if long pressed not used
      }
      break;

    case PRESSED:
      if (evt == EVT_INPUT_CHANGED) {
        if (jiffies - input->lastTime < DEBOUNCE) {
          JITTER_ADD(jiffies - input->lastTime, input);
          break; // ignoring glitch
        }
        //printk(KERN_INFO "Tobe-Pilot: Input %s pressed [%lu].\n", input->hw->name, jiffies);
        // Jitter on input release (before polling task caught it=> run pressed)
        input->lastTime = jiffies;
        input->state = IDLE;
        PRESSED_INC(input);
        if (input->hw->longPressed)
          inScriptRun(input->hw->pressed, 0);
      }

      else if (evt == EVT_TICK) {
        //printk(KERN_INFO "Tobe-Pilot: Tick %u.\n", jiffies);
        if (gpio_get_value(input->hw->gpio.pin)) { // Button has been released
          //printk(KERN_INFO "Tobe-Pilot: Input %s pressed [%lu].\n", input->hw->name, jiffies);
          input->lastTime = jiffies;
          input->state = IDLE;
          PRESSED_INC(input);
          if (input->hw->longPressed)
            inScriptRun(input->hw->pressed, 0);
        } else if (jiffies - input->lastTime > LONG_PRESS_TIME) {
          //printk(KERN_INFO "Tobe-Pilot: Input %s long pressed [%lu].\n", input->hw->name, jiffies);
          input->state = LONG_PRESSED;
          LONGPRESSED_INC(input);
          if (input->hw->longPressed)
            inScriptRun(input->hw->longPressed, 0);
        }
      }
      break;

    case LONG_PRESSED:
      if ( ((evt == EVT_TICK) && gpio_get_value(input->hw->gpio.pin)) ||
            (evt == EVT_INPUT_CHANGED) ) { // Button has been released
        input->state = IDLE;
        input->lastTime = jiffies;
      }
      break;
  }
}

// interrupts -----------------------------------------------------------------
static irqreturn_t inChangedIrq(int irq, void *dev) {
  //printk(KERN_INFO "Tobe-Pilot: input changed [%u].\n", jiffies);
  inSendEvent((Input *)dev, EVT_INPUT_CHANGED);
  return IRQ_HANDLED;
}

static void inPoll(void) {
  Input *ptr = inList;

  while (ptr) {
    if (ptr->state != IDLE)
      inSendEvent(ptr, EVT_TICK);
    ptr = ptr->next;
  }
}

static const char *envp[] = { "HOME=/home/pi", "TERM=linux", "PATH=/bin:/usr/bin:/usr/local/bin", NULL };

void inScriptRun(const char *script, int wait) {
  const char *argv[] = { "/bin/sh", "-c", script, NULL };

  //printk(KERN_INFO "ToBe-Pilot(script): Running %s\n", script);
  call_usermodehelper(argv[0], (char**)argv, (char**)envp, wait ? UMH_WAIT_PROC : UMH_NO_WAIT);
}


/** Setup a specific Input.
 * Note: inList is not synchronised, assuming all init is done at insmod...
 *       ... if not add spinlock.
 */
//void inAdd(unsigned gpioNb, const char *name, InputAction pressed, InputAction longPressed) {
void inAdd(const HwInput *hwIn) {
  Input *inp = kmalloc(sizeof(struct Input), GFP_KERNEL);
  if (!inp) return;
  inp->hw = hwIn;
  inp->state = IDLE;
  inp->lastTime = 0;
  inp->next = inList;
  inList = inp;
  inResetStatsForInput(inp);

  if (gpio_request(hwIn->gpio.pin, hwIn->name))
    printk(KERN_WARNING "ToBe-Pilot: Can't request GPIO %s ...\n", hwIn->name);

  if (request_irq(gpio_to_irq(hwIn->gpio.pin), inChangedIrq, IRQF_SHARED | IRQF_TRIGGER_FALLING, hwIn->name, inp))
    printk(KERN_WARNING "ToBe-Pilot: Can't request GPIO %s intr...\n", hwIn->name);
}

/** Initialize input handling. Setup pull ups and monitoring timer.
 * @param gpioMask enables Gpio <x> Pull Up when bit <x> is set (for GPIO 0..31).
 * @return status (0 = success)
 */
int inInit(unsigned pullupsMask) {
#ifdef __arm__
  struct GpioRegs *gpio;

  // Map GPIO registers (different location on 2835 -RPi1/Zero- and 2836/7 -RPi2/3-)
  gpio = (struct GpioRegs *)ioremap(rpiIsBCM2835() ? GPIO_BASE_BCM2835 : GPIO_BASE_BCM2836, 3);
  if (gpio == NULL) {
    return -1;
  }
  // Pull up configuration
  gpio->gppud = 2;
  ndelay(300); // Accoding to Datasheet, need to wait >150 cycles
  gpio->gppudclk[0] = pullupsMask;
  gpio->gppudclk[1] = 0;
  ndelay(300);
  gpio->gppudclk[0] = 0;
  iounmap(gpio);
#endif /* __arm__ */

  // start input polling timer
  tickAdd(inPoll);

  return 0;
}

/** Ends push button handling. */
void inExit(void) {
  tickDel(inPoll);

  while (inList) {
    Input *ptr = inList;
    inList = inList->next;
    free_irq(gpio_to_irq(ptr->hw->gpio.pin), ptr);
    gpio_free(ptr->hw->gpio.pin);
    kfree(ptr);
  }
}
