#include <linux/gpio.h>
#include <linux/string.h>
#include "pilot.h"
#include "tick.h"
#include "output.h"

typedef struct Output Output;

struct Output {
  struct device_attribute devAttr;
  const HwOutput *hw;
  struct {
    LedState state;
    int blinkPeriod, blinkDuty, flashDuration, cnt; // in 100ms ticks
  } led;
  Output *next;
};

static Output *outList = NULL;

// ============================================================================
static const char * getState(const HwOutput *hw) {
  unsigned val = gpio_get_value(hw->gpio.pin);
  if (hw->gpio.polarity == ACTIVE_LOW) val = (val ? 0 : 1);

  switch (hw->type) {
    case OUT_PWR:  return (val ? "on"    : "off");
    case OUT_RST:  return (val ? "reset" : "running");
    case OUT_BOOT: return (val ? "bootl" : "normal");
    case OUT_LED:  return (val ? "on"    : "off");
  }
  return "???";
}

static ssize_t show(struct device *dev, struct device_attribute *attr, char *buf) {
  Output *output  = container_of(attr, Output, devAttr);
  return scnprintf(buf, PAGE_SIZE, "%s\n", getState(output->hw));
}

static ssize_t store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
  Output *output  = container_of(attr, Output, devAttr);
  unsigned pin = output->hw->gpio.pin;

  switch (output->hw->type) {
    case OUT_PWR:
      if (sysfs_streq(buf, "on"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 0 : 1);
      else if (sysfs_streq(buf, "off"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 1 : 0);
      else if (sysfs_streq(buf, "toggle"))
        gpio_set_value(pin, gpio_get_value(pin) ? 0 : 1);
      break;

    case OUT_RST:
      if (sysfs_streq(buf, "on") || sysfs_streq(buf, "run"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 1 : 0);
      else if (sysfs_streq(buf, "reset"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 0 : 1);
      break;

    case OUT_BOOT:
      if (sysfs_streq(buf, "bootl"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 0 : 1);
      else if (sysfs_streq(buf, "normal"))
        gpio_set_value(pin, output->hw->gpio.polarity == ACTIVE_LOW ? 1 : 0);
      break;

    case OUT_LED:
      // TBD: right now via done status ...
      break;

    default:
      printk(KERN_WARNING "ToBe-Pilot-Out: Unknown type.\n");
  }
  return count;
}

static int outPrintStatsForOutput(char *buf, size_t size, Output *outp) {
  int nb;

  if (outp == NULL) return 0;

  nb = outPrintStatsForOutput(buf, size, outp->next);
  nb += scnprintf(buf+nb, size-nb, "  %-5s: %s\n", outp->hw->name, getState(outp->hw));
  return nb;
}

int outPrintStats(char *buf, size_t size) {
  int nb = scnprintf(buf, size, "Outputs:\n");
  return nb+outPrintStatsForOutput(buf+nb, size-nb, outList);
}

// ============================================================================

// this function is called every 100ms by the tick timer
static void ledPoll(void) {
  Output *ptr = outList;
  while (ptr) {
    switch (ptr->led.state) {
      case TOGGLE:
        if (ptr->led.state == OFF) {
          ptr->led.state = ON;
          gpio_set_value(ptr->hw->gpio.pin, 1);
        } else {
          ptr->led.state = OFF;
          gpio_set_value(ptr->hw->gpio.pin, 0);
        }
        break;

      case BLINK:
        gpio_set_value(ptr->hw->gpio.pin, ptr->led.cnt >= ptr->led.blinkDuty ? 0 : 1);
        (ptr->led.cnt)++;
        if (ptr->led.cnt >= ptr->led.blinkPeriod) ptr->led.cnt = 0;
        break;

      case FLASH_ONCE:
        (ptr->led.cnt)++;
        if (ptr->led.cnt >= ptr->led.flashDuration)  {
          ptr->led.state = OFF;
          gpio_set_value(ptr->hw->gpio.pin, 0);
        }

      default: break;
    }
    ptr = ptr->next;
  }
}

//static const char* STATES[] = { "OFF", "ON", "TOGGLE", "BLINK", "FLASH_ONCE" };
void outLedSet(const char *led, LedState s) {
  Output *ptr = outList;

  //printk(KERN_INFO "ToBe-Pilot: LED %s => %s\n", led, STATES[s]);
  while (ptr) {
    if (strcmp(led, ptr->hw->name) == 0) {
      ptr->led.cnt = 0;
      switch (s) {
        case OFF:
          gpio_set_value(ptr->hw->gpio.pin, 0);
          ptr->led.state = OFF;
          break;

        case TOGGLE:
          if (ptr->led.state == OFF) {
            gpio_set_value(ptr->hw->gpio.pin, 1);
            ptr->led.state = ON;
          } else {
            gpio_set_value(ptr->hw->gpio.pin, 0);
            ptr->led.state = OFF;
          }
          break;

        default: // ON, BLINK, FLASH_ONCE
          gpio_set_value(ptr->hw->gpio.pin, 1);
          ptr->led.state = s;
          break;
      }
    }
    ptr = ptr->next;
  }
}

// ============================================================================

void outAdd(const HwOutput *hw) {
  unsigned pin = hw->gpio.pin;
  Output *output = kmalloc(sizeof(Output), GFP_KERNEL);

  if (!output) {
    printk(KERN_ERR "ToBe-Pilot-Out: Can't kmalloc.\n");
    return;
  }

  output->devAttr.attr.name = hw->name;
  output->devAttr.attr.mode = S_IWUSR | S_IRUGO;
  output->devAttr.show = show;
  output->devAttr.store = store;
  output->hw = hw;
  output->next = outList;
  outList = output;

  if (gpio_request(pin, hw->name)) {
    printk(KERN_ERR "ToBe-Pilot-Out: Can't request GPIO '%s' (%d).\n", hw->name, pin);
    return;
  }

  switch (hw->type) {
    case OUT_PWR:
    case OUT_RST:
    case OUT_BOOT:
      gpio_direction_output(pin, gpio_get_value(pin)); // keeps default state
      break;

    case OUT_LED:
      output->led.state = OFF;
      output->led.blinkPeriod = 4;
      output->led.blinkDuty = 1;
      output->led.flashDuration = 15;
      output->led.cnt = 0;

      gpio_direction_output(pin, hw->gpio.polarity == ACTIVE_HIGH ? 0 : 1); // LED off
      break;
  }
}

void outCreateSysFs(struct kobject *obj) {
  Output *ptr = outList;
  while (ptr) {
    if (sysfs_create_file(obj, &(ptr->devAttr.attr)))
      printk(KERN_WARNING "ToBe-Pilot-Out: Can't create sysfs for %s.\n", ptr->hw->name);
    ptr = ptr->next;
  }
}

void outInit(void) {
  tickAdd(ledPoll);
}

void outExit(struct kobject *obj) {
  tickDel(ledPoll);

  while (outList) {
    Output *ptr = outList;
    outList = outList->next;
    sysfs_remove_file(obj, &(ptr->devAttr.attr));
    gpio_free(ptr->hw->gpio.pin);
    kfree(ptr);
  }
}
